import { FormGroup, FormArray } from '@angular/forms';
import { from, forkJoin } from 'rxjs';
import { map } from 'rxjs/operators';

function lazy(factory, destroy) {
    let init = false;
    let lVal;
    let err;
    function value() {
        if (!init) {
            init = true;
            try {
                lVal = factory();
            }
            catch (e) {
                err = e;
            }
        }
        if (err) {
            throw err;
        }
        return lVal;
    }
    value.reset = () => {
        if (init) {
            err = undefined;
            if (typeof destroy === 'function') {
                destroy(lVal);
            }
            init = false;
        }
    };
    value.isInit = () => init;
    return value;
}

function traverseControl(control, ...ps) {
    if (ps.length === 0 || !control) {
        return control;
    }
    if (control instanceof FormGroup || control instanceof FormArray) {
        const [p, ...lps] = ps;
        return traverseControl(control.controls[p], ...lps);
    }
}

function getElements(ctrls) {
    return Array.isArray(ctrls) ? ctrls :
        typeof ctrls === 'object' && Object.values(ctrls) || [];
}
function tapControls(root, action, updateState) {
    for (const ctrl of getElements(root && root.controls)) {
        if (ctrl instanceof FormGroup || ctrl instanceof FormArray) {
            tapControls(ctrl, action, updateState);
            continue;
        }
        if (typeof action === 'function') {
            action(ctrl);
        }
        if (updateState) {
            ctrl.updateValueAndValidity();
        }
    }
}

function mergeError(errorArray) {
    const sanitizedErrors = errorArray === null || errorArray === void 0 ? void 0 : errorArray.filter(({ err }) => Object.keys(err || {}).length);
    if (sanitizedErrors === null || sanitizedErrors === void 0 ? void 0 : sanitizedErrors.length) {
        return Object.assign({}, ...sanitizedErrors.map(({ key, err }) => ({ [key]: err })));
    }
}
function loopControl(control, selector) {
    if (control instanceof FormGroup) {
        return selector(Object.entries(control.controls)
            .map(([key, ctrl]) => ({ key, control: ctrl })));
    }
    else if (control instanceof FormArray) {
        return selector(control.controls
            .map((ctrl, key) => ({ key, control: ctrl })));
    }
}
function compositeValidator() {
    const validator = (ctrl) => {
        const vl = ctrl === null || ctrl === void 0 ? void 0 : ctrl.validator;
        if (typeof vl === 'function') {
            return vl(ctrl);
        }
        const errAry = loopControl(ctrl, pairs => pairs.reduce((acc, { key, control }) => [...acc, { key, err: validator(control) }], []));
        return mergeError(errAry);
    };
    return control => {
        if (control instanceof FormGroup || control instanceof FormArray) {
            return validator(control);
        }
    };
}
function compositeAsyncValidator() {
    const validator = (ctrl) => {
        const vl = ctrl === null || ctrl === void 0 ? void 0 : ctrl.asyncValidator;
        if (typeof vl === 'function') {
            return vl(ctrl);
        }
        const asyncErrAry = loopControl(ctrl, pairs => pairs.reduce((acc, { key, control }) => {
            const asyncErr = validator(control);
            if (asyncErr) {
                return [...acc, from(asyncErr).pipe(map(err => ({ key, err })))];
            }
            return acc;
        }, []));
        if (asyncErrAry) {
            return forkJoin(asyncErrAry).pipe(map(errAry => mergeError(errAry)));
        }
    };
    return control => {
        if (control instanceof FormGroup || control instanceof FormArray) {
            return validator(control);
        }
    };
}

function uniq(items) {
    return (items || []).filter((val, idx, self) => val && self.indexOf(val) === idx);
}

/**
 * Generated bundle index. Do not edit.
 */

export { compositeAsyncValidator, compositeValidator, lazy, tapControls, traverseControl, uniq };
//# sourceMappingURL=ngx-reactive-form-utils.js.map
